/* Copyright (C) 2023-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: S.Gorbunov[committer], N.Bluhme */

/// @file    CbmBbaAlignmentTask.cxx
/// @author  Sergey Gorbunov
/// @author  Nora Bluhme
/// @date    20.01.2023
/// @brief   Task class for alignment
///

// Cbm Headers ----------------------
#include "CbmBbaAlignmentTask.h"


// ROOT headers

#include "FairRootManager.h"
#include "TClonesArray.h"
#include "TFile.h"
#include "TH1F.h"
#include "TRandom.h"

//
#include "CbmCaTimeSliceReader.h"
#include "CbmGlobalTrack.h"
#include "CbmL1.h"
#include "CbmMuchTrack.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdHit.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmStsHit.h"
#include "CbmStsSetup.h"
#include "CbmStsTrack.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofTrack.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdTrack.h"
#include "CbmTrdTrackingInterface.h"
#include "TCanvas.h"
#include "bba/BBA.h"

#include <cmath>

// c++ and std headers

#include <iostream>

namespace
{
  using namespace cbm::algo;
}

void CbmBbaAlignmentTask::TrackContainer::MakeConsistent()
{
  fNmvdHits  = 0;
  fNstsHits  = 0;
  fNmuchHits = 0;
  fNtrdHits  = 0;
  fNtofHits  = 0;
  for (const auto& n : fUnalignedTrack.fNodes) {
    switch (n.fHitSystemId) {
      case ECbmModuleId::kMvd: fNmvdHits++; break;
      case ECbmModuleId::kSts: fNstsHits++; break;
      case ECbmModuleId::kMuch: fNmuchHits++; break;
      case ECbmModuleId::kTrd: fNtrdHits++; break;
      case ECbmModuleId::kTof: fNtofHits++; break;
      default: break;
    }
  }
}


CbmBbaAlignmentTask::CbmBbaAlignmentTask(const char* name, Int_t iVerbose, TString histoFileName)
  : FairTask(name, iVerbose)
  , fHistoFileName(histoFileName)
{
  TFile* curFile           = gFile;
  TDirectory* curDirectory = gDirectory;

  if (!(fHistoFileName == "")) {
    fHistoFile = new TFile(fHistoFileName.Data(), "RECREATE");
  }
  else {
    fHistoFile = gFile;
  }

  fHistoFile->cd();

  fHistoDir = fHistoFile->mkdir("Bba");
  fHistoDir->cd();

  gFile      = curFile;
  gDirectory = curDirectory;
}


CbmBbaAlignmentTask::~CbmBbaAlignmentTask() {}


InitStatus CbmBbaAlignmentTask::Init()
{

  fFitter.Init();

  //Get ROOT Manager
  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(error) << "CbmBbaAlignmentTask::Init :: RootManager not instantiated!";
    return kERROR;
  }

  // Get global tracks & matches

  if (TrackingMode::kSts != fTrackingMode) {

    fInputGlobalTracks = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));
    if (!fInputGlobalTracks) {
      LOG(error) << "CbmBbaAlignmentTask::Init: Global track array not found!";
      return kERROR;
    }

    fInputGlobalTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrackMatch"));

    if (!fInputGlobalTrackMatches) {
      LOG(error) << "CbmBbaAlignmentTask::Init: Global track matches array not found!";
      //return kERROR;
    }
  }

  fInputStsTracks = static_cast<TClonesArray*>(ioman->GetObject("StsTrack"));
  if (!fInputStsTracks) {
    LOG(error) << "CbmBbaAlignmentTask::Init: Sts track array not found!";
    return kERROR;
  }

  fInputStsTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("StsTrackMatch"));

  if (!fInputStsTrackMatches) {
    LOG(error) << "CbmBbaAlignmentTask::Init: Sts track matches array not found!";
    //return kERROR;
  }

  // MC tracks
  fInputMcTracks = static_cast<TClonesArray*>(ioman->GetObject("MCTrack"));
  if (!fInputMcTracks) {
    Warning("CbmBbaAlignmentTask::Init", "mc track array not found!");
    return kERROR;
  }

  fTracks.clear();


  ca::Framework* l1                       = CbmL1::Instance()->fpAlgo;
  const ca::Parameters<ca::fvec>& l1Param = l1->GetParameters();
  fNalignmentBodies                       = l1Param.GetNstationsActive();

  TDirectory* curDirectory = gDirectory;

  fHistoDir->cd();

  for (int i = -1; i < fNalignmentBodies; i++) {
    const char* n1 = Form("Body%d", i);
    const char* n2 = Form(", body %d", i);
    if (i == -1) {
      n1 = "All";
      n2 = "";
    }
    fHistoDir->mkdir(n1);
    fHistoDir->cd(n1);

    hResidualsBeforeAlignmentX.push_back(
      new TH1F(Form("ResBeforeAli%s_X", n1), Form("X-Residuals Before Alignment%s", n2), 100, -1.5, 1.5));
    hResidualsBeforeAlignmentY.push_back(
      new TH1F(Form("ResBeforeAli%s_Y", n1), Form("Y-Residuals Before Alignment%s", n2), 100, -1.5, 1.5));
    hResidualsAfterAlignmentX.push_back(
      new TH1F(Form("ResAfterAli%s_X", n1), Form("X-Residuals After Alignment%s", n2), 100, -1.5, 1.5));
    hResidualsAfterAlignmentY.push_back(
      new TH1F(Form("ResAfterAli%s_Y", n1), Form("Y-Residuals After Alignment%s", n2), 100, -1.5, 1.5));

    hPullsBeforeAlignmentX.push_back(
      new TH1F(Form("PullBeforeAli%s_X", n1), Form("X-Pulls Before Alignment%s", n2), 100, -10., 10.));
    hPullsBeforeAlignmentY.push_back(
      new TH1F(Form("PullBeforeAli%s_Y", n1), Form("Y-Pulls Before Alignment%s", n2), 100, -10., 10.));
    hPullsAfterAlignmentX.push_back(
      new TH1F(Form("PullAfterAli%s_X", n1), Form("X-Pulls After Alignment%s", n2), 100, -10., 10.));
    hPullsAfterAlignmentY.push_back(
      new TH1F(Form("PullAfterAli%s_Y", n1), Form("Y-Pulls After Alignment%s", n2), 100, -10., 10.));
  }

  gDirectory = curDirectory;

  return kSUCCESS;
}

void CbmBbaAlignmentTask::Exec(Option_t* /*opt*/)
{

  LOG(info) << "BBA: exec event N " << fNEvents;

  fNEvents++;

  if (static_cast<int>(fTracks.size()) >= fMaxNtracks) {
    return;
  }

  ca::Framework* l1                     = CbmL1::Instance()->fpAlgo;
  const ca::Parameters<ca::fvec>& l1Par = l1->GetParameters();

  static int statGlobalTracks = 0;
  statGlobalTracks += fInputGlobalTracks->GetEntriesFast();

  // select tracks for alignment and store them
  if (fTrackingMode == kSts && fInputStsTracks) {
    for (int iTr = 0; iTr < fInputStsTracks->GetEntriesFast(); iTr++) {
      if (static_cast<int>(fTracks.size()) >= fMaxNtracks) {
        break;
      }
      const CbmStsTrack* stsTrack = dynamic_cast<CbmStsTrack*>(fInputStsTracks->At(iTr));
      if (!stsTrack) continue;
      TrackContainer t;
      if (!fFitter.CreateMvdStsTrack(t.fUnalignedTrack, iTr)) continue;
      t.MakeConsistent();
      fFitter.FitTrack(t.fUnalignedTrack);
      t.fAlignedTrack = t.fUnalignedTrack;
      if (t.fNstsHits < l1Par.GetNstationsActive(ca::EDetectorID::kSts)) continue;
      const auto& par = t.fUnalignedTrack.fNodes[t.fUnalignedTrack.fFirstHitNode].fTrack;
      if (!std::isfinite((ca::fscal) par.GetQp()[0])) continue;
      if (fabs(par.GetQp()[0]) > 1.) continue;  // select tracks with min 1 GeV momentum
      fTracks.push_back(t);
    }
  }
  else if (fTrackingMode == kMcbm && fInputGlobalTracks) {
    for (int iTr = 0; iTr < fInputGlobalTracks->GetEntriesFast(); iTr++) {
      if (static_cast<int>(fTracks.size()) >= fMaxNtracks) {
        break;
      }
      const CbmGlobalTrack* globalTrack = dynamic_cast<const CbmGlobalTrack*>(fInputGlobalTracks->At(iTr));
      if (!globalTrack) {
        LOG(fatal) << "BBA: null pointer to the global track!";
        break;
      }
      TrackContainer t;
      if (!fFitter.CreateGlobalTrack(t.fUnalignedTrack, *globalTrack)) {
        LOG(fatal) << "BBA: can not create the global track for the fit! ";
        break;
      }
      t.MakeConsistent();
      t.fUnalignedTrack.fMsQp0                                              = 1. / 0.1;
      t.fUnalignedTrack.fIsMsQp0Set                                         = true;
      t.fUnalignedTrack.fNodes[t.fUnalignedTrack.fFirstHitNode].fTrack.Qp() = 0.;

      fFitter.FitTrack(t.fUnalignedTrack);
      t.fAlignedTrack = t.fUnalignedTrack;
      //if (t.fNstsHits < 1) continue;
      //if (t.fNtrdHits < 2) continue;
      if (t.fNstsHits + t.fNtrdHits + t.fNtofHits < 4) continue;
      fTracks.push_back(t);
    }
  }

  LOG(info) << "Bba: " << fInputGlobalTracks->GetEntriesFast() << " global tracks in event, " << statGlobalTracks
            << " global tracks in total, " << fTracks.size() << " tracks selected for alignment";
}


void CbmBbaAlignmentTask::ApplyAlignment(const std::vector<double>& par)
{
  // apply alignment parameters to the hits

  for (TrackContainer& t : fTracks) {

    for (unsigned int in = 0; in < t.fUnalignedTrack.fNodes.size(); in++) {
      CbmKfTrackFitter::FitNode& node        = t.fUnalignedTrack.fNodes[in];
      CbmKfTrackFitter::FitNode& nodeAligned = t.fAlignedTrack.fNodes[in];
      int iAlignmentModule                   = node.fReference1;
      if (iAlignmentModule < 0) continue;
      assert(nodeAligned.fReference1 == iAlignmentModule);

      double dx = par[3 * iAlignmentModule + 0];
      double dy = par[3 * iAlignmentModule + 1];
      double dz = par[3 * iAlignmentModule + 2];

      nodeAligned.fMxy.SetX(node.fMxy.X() + ca::fvec(dx));
      nodeAligned.fMxy.SetY(node.fMxy.Y() + ca::fvec(dy));
      nodeAligned.fZ = node.fZ + dz;
    }
  }
}


double CbmBbaAlignmentTask::CostFunction(const std::vector<double>& par)
{
  // apply new parameters to the hits

  for (auto& t : fTracks) {
    t.fAlignedTrack = t.fUnalignedTrack;
  }

  ApplyAlignment(par);

  fChi2Total = 0.;
  fNdfTotal  = 0;

  for (auto& t : fTracks) {
    if (!t.fIsActive) continue;
    fFitter.FitTrack(t.fAlignedTrack);
    double chi2 = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetChiSq()[0];
    double ndf  = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetNdf()[0];
    if (!std::isfinite(chi2) || chi2 <= 0. || ndf <= 0.) {
      chi2 = 0.;
      ndf  = 0.;
    }
    fChi2Total += chi2;
    fNdfTotal += (int) ndf;
  }

  double cost = fChi2Total / (fNdfTotal + 1);
  if (fFixedNdf > 0 && fNdfTotal != fFixedNdf) {
    cost = 1.e30;
  }
  LOG(info) << "BBA: cost function:  ndf " << fNdfTotal << ", cost " << cost
            << ", diff to ideal cost: " << cost - fCostIdeal;

  return cost;
}

void CbmBbaAlignmentTask::Finish()
{
  //
  // perform the alignment
  //

  LOG(info) << "BBA: start the alignment procedure with " << fTracks.size() << " tracks ...";

  for (auto& t : fTracks) {
    for (auto& n : t.fUnalignedTrack.fNodes) {
      // TODO: get the body index from n.fHitSystemId, n.fHitAddress
      int iBody     = n.fMaterialLayer;  // material layer is currently equal to the active tracking station
      n.fReference1 = iBody;             // reference1 will be the body index
    }
    t.fAlignedTrack = t.fUnalignedTrack;
  }

  int nParameters = 3 * fNalignmentBodies;  // x, y, z

  std::vector<bba::Parameter> par(nParameters);

  for (int ip = 0; ip < nParameters; ip++) {
    bba::Parameter& p = par[ip];
    p.SetActive(0);
    p.SetValue(0.);
    p.SetBoundaryMin(-4.);  // +-1 cm alignment range
    p.SetBoundaryMax(4.);
    p.SetStepMin(1.e-4);
    p.SetStepMax(0.1);
    p.SetStep(1.e-2);
  }

  par[0].SetActive(0);  // fix the first station
  par[1].SetActive(0);
  par[2].SetActive(0);

  par[3 * (fNalignmentBodies - 1) + 0].SetActive(0);  // fix the last station
  par[3 * (fNalignmentBodies - 1) + 1].SetActive(0);
  par[3 * (fNalignmentBodies - 1) + 2].SetActive(0);

  // set active parameters for other stations

  for (int i = 1; i < fNalignmentBodies - 1; i++) {
    par[3 * i + 0].SetActive(0);
    par[3 * i + 1].SetActive(0);
    par[3 * i + 2].SetActive(0);
  }

  for (int i = 1; i < fNalignmentBodies - 1; i++) {
    //for (int i = 1; i < 2; i++) {
    par[3 * i + 0].SetActive(1);
    par[3 * i + 1].SetActive(1);
    par[3 * i + 2].SetActive(1);
  }

  gRandom->SetSeed(1);

  for (int is = 0; is < fNalignmentBodies; is++) {
    bba::Parameter& px = par[3 * is + 0];
    bba::Parameter& py = par[3 * is + 1];
    bba::Parameter& pz = par[3 * is + 2];

    // +- 0.5 cm misalignment
    double d = .2;
    if (px.IsActive()) {
      px.SetValue(gRandom->Uniform(2. * d) - d);
    }
    if (py.IsActive()) {
      py.SetValue(gRandom->Uniform(2. * d) - d);
    }
    if (pz.IsActive()) {
      pz.SetValue(gRandom->Uniform(2. * d) - d);
    }
  }

  bba::BBA alignment;

  alignment.setParameters(par);

  auto lambdaCostFunction = [this](const std::vector<double>& p) { return CostFunction(p); };

  alignment.setChi2(lambdaCostFunction);
  alignment.printInfo();

  std::vector<double> parIdeal;
  std::vector<double> parMisaligned;
  {
    const std::vector<double>& r = alignment.getResult();
    for (unsigned int i = 0; i < r.size(); i++) {
      parMisaligned.push_back(r[i]);
      parIdeal.push_back(0.);
    }
  }

  {
    std::cout << "initial cost function..." << std::endl;

    fCostInitial = CostFunction(parMisaligned);

    for (auto& t : fTracks) {
      double ndf  = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetNdf()[0];
      double chi2 = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetChiSq()[0];
      if (ndf < 0. || chi2 < 0. || !std::isfinite(ndf) || !std::isfinite(chi2)) {
        t.fIsActive = false;
      }
    }
    std::cout << "recalculate initial cost function..." << std::endl;
    fCostInitial = CostFunction(parMisaligned);
    fFixedNdf    = -1;  //fNdfTotal;


    for (auto& t : fTracks) {
      std::cout << "THIS IS THE HISTOGRAM LOOP" << std::endl;
      if (!t.fIsActive) continue;
      // calculate the residuals for all tracks at each fitNode before alignment
      for (unsigned int in = 0; in < t.fAlignedTrack.fNodes.size(); in++) {
        CbmKfTrackFitter::FitNode& node = t.fAlignedTrack.fNodes[in];

        if (!node.fIsFitted || !node.fIsXySet) continue;

        double x_residual = node.fMxy.X()[0] - node.fTrack.X()[0];  // this should be the difference vector
        double y_residual = node.fMxy.Y()[0] - node.fTrack.Y()[0];
        hResidualsBeforeAlignmentX[0]->Fill(x_residual);
        hResidualsBeforeAlignmentY[0]->Fill(y_residual);
        hResidualsBeforeAlignmentX[node.fReference1 + 1]->Fill(x_residual);
        hResidualsBeforeAlignmentY[node.fReference1 + 1]->Fill(y_residual);

        double x_pull = x_residual / sqrt(node.fMxy.Dx2()[0] + node.fTrack.GetCovariance(0, 0)[0]);
        double y_pull = y_residual / sqrt(node.fMxy.Dy2()[0] + node.fTrack.GetCovariance(1, 1)[0]);

        hPullsBeforeAlignmentX[0]->Fill(x_pull);
        hPullsBeforeAlignmentY[0]->Fill(y_pull);
        hPullsBeforeAlignmentX[node.fReference1 + 1]->Fill(x_pull);
        hPullsBeforeAlignmentY[node.fReference1 + 1]->Fill(y_pull);
      }
    }
  }

  std::cout << "ideal cost function..." << std::endl;
  fCostIdeal = CostFunction(parIdeal);


  LOG(info) << "Initial nTracks: " << fTracks.size() << " cost " << fCostInitial;
  LOG(info) << " cost function for the true parameters: " << fCostIdeal;

  alignment.align();

  LOG(info) << " cost function for the true parameters: " << fCostIdeal;

  LOG(info) << " Misaligned parameters: ";
  for (int is = 0; is < fNalignmentBodies; is++) {
    const std::vector<double>& r = parMisaligned;
    LOG(info) << "Sts station " << is << ": x " << r[3 * is + 0] << " y " << r[3 * is + 1] << " z " << r[3 * is + 2];
  }

  LOG(info) << " Alignment results: ";

  for (int is = 0; is < fNalignmentBodies; is++) {
    const std::vector<double>& r = alignment.getResult();
    LOG(info) << "Sts station " << is << ": x " << r[3 * is + 0] << " y " << r[3 * is + 1] << " z " << r[3 * is + 2];
  }

  CostFunction(alignment.getResult());

  for (auto& t : fTracks) {
    if (!t.fIsActive) continue;
    // calculate the residuals for all tracks at each fitNode before alignment
    for (unsigned int in = 0; in < t.fAlignedTrack.fNodes.size(); in++) {
      CbmKfTrackFitter::FitNode& node = t.fAlignedTrack.fNodes[in];

      if (!node.fIsFitted || !node.fIsXySet) continue;

      double x_residual = node.fMxy.X()[0] - node.fTrack.X()[0];
      double y_residual = node.fMxy.Y()[0] - node.fTrack.Y()[0];
      hResidualsAfterAlignmentX[0]->Fill(x_residual);
      hResidualsAfterAlignmentY[0]->Fill(y_residual);
      hResidualsAfterAlignmentX[node.fReference1 + 1]->Fill(x_residual);
      hResidualsAfterAlignmentY[node.fReference1 + 1]->Fill(y_residual);

      hPullsAfterAlignmentX[0]->Fill(x_residual);
      hPullsAfterAlignmentY[0]->Fill(y_residual);
      hPullsAfterAlignmentX[node.fReference1 + 1]->Fill(x_residual);
      hPullsAfterAlignmentY[node.fReference1 + 1]->Fill(y_residual);
    }
  }

  // store the histograms

  TDirectory* curr   = gDirectory;
  TFile* currentFile = gFile;
  // Open output file and write histograms

  fHistoFile->cd();
  WriteHistosCurFile(fHistoDir);
  if (!(fHistoFileName == "")) {
    fHistoFile->Close();
    fHistoFile->Delete();
  }
  gFile      = currentFile;
  gDirectory = curr;
}

void CbmBbaAlignmentTask::WriteHistosCurFile(TObject* obj)
{
  if (!obj->IsFolder())
    obj->Write();
  else {
    TDirectory* cur    = gDirectory;
    TFile* currentFile = gFile;

    TDirectory* sub = cur->GetDirectory(obj->GetName());
    sub->cd();
    TList* listSub = (static_cast<TDirectory*>(obj))->GetList();
    TIter it(listSub);
    while (TObject* obj1 = it()) {
      WriteHistosCurFile(obj1);
    }
    cur->cd();
    gFile      = currentFile;
    gDirectory = cur;
  }
}

ClassImp(CbmBbaAlignmentTask);
